import inspect
import markdown
from typing import Callable, List, Type, Union

from nio import Event, InviteEvent

from core.bot_interface import BotInterface
from core.command import Command
from core.logger import logger


class BaseModule:
    def __init__(self, bot: BotInterface):
        self.bot = bot
        self._pending_subcommands: List[Callable] = []
        self.register_handlers()

    def on_start(self):
        pass

    def on_stop(self):
        pass

    def register_handlers(self):
        for name, method in inspect.getmembers(self, predicate=inspect.iscoroutinefunction):
            if hasattr(method, "_command_info"):
                self._register_command(method)
            elif hasattr(method, "_subcommand_info"):
                self._pending_subcommands.append(method)
            elif hasattr(method, "_event_type"):
                self._register_event_handler(method)

        for method in self._pending_subcommands:
            self._register_subcommand(method)

    def _register_command(self, method: Callable):
        command_info = getattr(method, "_command_info")
        required_level = getattr(method, "_required_level", 0)
        command = Command(
            name=command_info["name"],
            description=command_info["description"],
            handler=method,
            required_level=required_level
        )
        self.bot.command_processor.register_command(command)

    def _register_subcommand(self, method: Callable):
        subcommand_info = getattr(method, "_subcommand_info")
        parent_command = self.bot.command_processor.commands.get(subcommand_info["parent_name"])

        if not parent_command:
            logger.warning(f"Parent command '{subcommand_info['parent_name']}' not found for subcommand '{subcommand_info['name']}'")
            return

        required_level = getattr(method, "_required_level", 0)
        subcommand = Command(
            name=subcommand_info["name"],
            description=subcommand_info["description"],
            handler=method,
            required_level=required_level
        )
        parent_command.add_subcommand(subcommand)

    def _register_event_handler(self, method: Callable):
        event_type = getattr(method, "_event_type")
        self.bot.event_dispatcher.subscribe(event_type, method)

    @classmethod
    def command_handler(cls, name: str, description: str, required_level: int = 0):
        def decorator(func: Callable):
            func._command_info = {"name": name, "description": description}
            func._required_level = required_level
            func.subcommand = cls._create_subcommand_decorator(name)
            return func

        return decorator

    @classmethod
    def _create_subcommand_decorator(cls, parent_name: str):
        def subcommand_decorator(name: str, description: str, required_level: int = 0):
            def decorator(func: Callable):
                func._subcommand_info = {"parent_name": parent_name, "name": name, "description": description}
                func._required_level = required_level
                return func

            return decorator

        return subcommand_decorator

    @classmethod
    def on_event(cls, event_type: Type[Union[Event, InviteEvent]]):
        def decorator(func: Callable):
            func._event_type = event_type
            return func

        return decorator

    async def send_message(self, room_id: str, message: str, send_formatted: bool = True):
        if send_formatted:
            formatted_body = {
                "msgtype": "m.text",
                "body": message,
                "format": "org.matrix.custom.html",
                "formatted_body": markdown.markdown(message)
            }
        else:
            formatted_body = {
                "msgtype": "m.text",
                "body": message
            }

        await self.bot.client.room_send(
            room_id,
            "m.room.message",
            formatted_body
        )
