from datetime import datetime
from typing import Union
from nio import AsyncClient, Event, InviteEvent, RoomMessageText, MatrixRoom
from core.bot_interface import BotInterface
from core.module_manager import ModuleManager
from core.command import PermissionManager, CommandProcessor
from core.event_dispatcher import EventDispatcher
from core.config import ConfigManager
from core.logger import logger


class Bot(BotInterface):
    def __init__(self, config: ConfigManager):
        self._config = config
        self._homeserver = self._config.get("credentials.homeserver")
        self._user_id = self._config.get("credentials.user_id")
        self._password = self._config.get("credentials.password")
        self._token = self._config.get("credentials.token")

        self._module_folder = self._config.get("bot.module_folder")
        self._command_prefix = self._config.get("bot.command_prefix")
        self._whitelisted_users = self._config.get("bot.whitelisted_users")

        self._client = AsyncClient(homeserver=self._homeserver, user=self._user_id)

        self._event_dispatcher = EventDispatcher()
        self._permission_manager = PermissionManager(self._client, self._whitelisted_users)
        self._command_processor = CommandProcessor(self._permission_manager)
        self.module_manager = ModuleManager(self)

        self._start_time = None

    async def start(self):
        logger.info("Bot started!")
        self._start_time = datetime.now()

        if self._token:
            self.client.access_token = self._token
        else:
            await self.client.login(self._password)

        self.module_manager.load_modules_from_folder(self._module_folder)
        self.client.add_event_callback(self._handle_event, Union[Event, InviteEvent])
        await self.client.sync_forever(timeout=30000)

    async def stop(self):
        for module in self.module_manager.modules:
            self.module_manager.unregister_module(module)
        await self._client.logout()
        logger.info("Bot stopped!")

    async def _handle_event(self, room: MatrixRoom, event: Union[Event, InviteEvent]):
        if isinstance(event, RoomMessageText):
            event_time = datetime.fromtimestamp(event.server_timestamp / 1000)
            if self._start_time and event_time < self._start_time:
                return

            if event.body.startswith(self._command_prefix):
                await self.command_processor.execute_command(room, event.body, event)

        await self.event_dispatcher.dispatch(room, event)

    @property
    def client(self) -> AsyncClient:
        return self._client

    @property
    def command_processor(self) -> CommandProcessor:
        return self._command_processor

    @property
    def event_dispatcher(self) -> EventDispatcher:
        return self._event_dispatcher

    @property
    def config(self) -> ConfigManager:
        return self._config

