from typing import List, Callable, Dict
from nio import MatrixRoom, Event, AsyncClient
from core.logger import logger

class PermissionManager:
    def __init__(self, client: AsyncClient, whitelisted_users: List[str]):
        self.client = client
        self.whitelisted_users = whitelisted_users

    async def get_user_power_level(self, room_id: str, user_id: str) -> int:
        try:
            power_levels = await self.client.room_get_state_event(room_id, "m.room.power_levels")
            users_power_levels = power_levels.content.get("users", {})
            return users_power_levels.get(user_id, 0)
        except Exception as e:
            logger.error(f"Failed to get power levels for room {room_id}: {e}")
            return 0

    async def has_permission(self, room_id: str, user_id: str, required_level: int) -> bool:
        if user_id in self.whitelisted_users:
            return True
        user_level = await self.get_user_power_level(room_id, user_id)
        return user_level >= required_level

class Command:
    def __init__(self, name: str, description: str, handler: Callable, required_level: int = 0):
        self.name = name
        self.description = description
        self.handler = handler
        self.subcommands: Dict[str, Command] = {}
        self.required_level = required_level

    def add_subcommand(self, subcommand) -> None:
        self.subcommands[subcommand.name] = subcommand

    async def execute(self, room, args: List[str], event: Event, permission_manager: PermissionManager) -> None:
        user_id = event.sender
        if not await permission_manager.has_permission(room.room_id, user_id, self.required_level):
            # await self.bot.send_message(room.room_id, "You do not have permission to execute this command.")
            return

        if args and args[0] in self.subcommands:
            subcommand = self.subcommands[args[0]]
            await subcommand.execute(room, args[1:], event, permission_manager)
        else:
            await self.handler(room, args, event)


class CommandProcessor:
    def __init__(self, permission_manager: PermissionManager):
        self.commands: Dict[str, Command] = {}
        self.permission_manager = permission_manager

    def register_command(self, command: Command):
        self.commands[command.name] = command

    async def execute_command(self, room: MatrixRoom, command_message: str, event: Event):
        parts = command_message.split(" ")
        if not parts:
            return

        command_message = parts[0][1:]
        if command_message in self.commands:
            command = self.commands[command_message]
            await command.execute(room, parts[1:], event, self.permission_manager)
        else:
            logger.warning(f"Command {command_message} not found")

    def get_commands(self) -> List[Dict[str, str]]:
        return [
            {
                "name": command.name,
                "description": command.description,
                "subcommands": [
                    {
                        "name": subcommand.name,
                        "description": subcommand.description,
                    }
                    for subcommand in command.subcommands.values()
                ],
            }
            for command in self.commands.values()
        ]
