import importlib
import inspect
import os
from pathlib import Path
from typing import List, Type

from core.base_module import BaseModule
from core.bot_interface import BotInterface
from core.logger import logger


class ModuleManager:
    def __init__(self, bot: BotInterface):
        self.modules = []
        self.module_folder = None
        self.bot: BotInterface = bot

    def load_modules_from_folder(self, module_folder: str):
        self.module_folder = module_folder
        loaded_modules = self.load_modules()
        for module_class in loaded_modules:
            self.register_module(module_class)

    def _load_module(self, module_name: str) -> Type['BaseModule']:
        module_path = os.path.join(self.module_folder, module_name)
        module_name_for_import = module_path.replace(os.sep, ".")

        try:
            module = importlib.import_module(module_name_for_import)
        except Exception as e:
            logger.error(f"Error importing module {module_name_for_import}: {e}")
            raise

        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and issubclass(obj, BaseModule) and obj != BaseModule:
                return obj
        raise ValueError(f"No valid BaseModule subclass found in module {module_name}")

    def load_modules(self) -> List[Type['BaseModule']]:
        modules = []
        for module_dir in Path(self.module_folder).iterdir():
            if module_dir.is_dir() and (module_dir / "__init__.py").exists():
                module_name = module_dir.name
                try:
                    module_class = self._load_module(module_name)
                    modules.append(module_class)
                except Exception as e:
                    logger.error(f"Failed to load module {module_name}: {e}")  # TODO: Proper logging
        return modules

    def register_module(self, module_class: Type['BaseModule']):
        try:
            module_instance = module_class(self.bot)
            self.modules.append(module_instance)
            module_instance.on_start()
            logger.info(f"Module {module_class.__name__} registered and started.")
        except Exception as e:
            logger.error(f"Failed to register module {module_class.__name__}: {e}")

    def unregister_module(self, module: 'BaseModule'):
        try:
            module.on_stop()
            self.modules.remove(module)
            logger.info(f"Module {module.__class__.__name__} unregistered and stopped.")
        except Exception as e:
            logger.error(f"Failed to unregister module {module.__class__.__name__}: {e}")
