import io
import os
import asyncio
import aiohttp
import json
from time import mktime
from typing import Dict, List, Optional, Any
from datetime import datetime

import feedparser
from nio import MatrixRoom, Event, RoomMessageText

from core.bot_interface import BotInterface
from core.base_module import BaseModule
from core.logger import logger
from core.config import ConfigManager


class RSSModule(BaseModule):
    def __init__(self, bot: BotInterface):
        super().__init__(bot)
        self._config = ConfigManager(os.path.join(os.path.dirname(__file__), "config.toml"))
        self._subscriptions_file = os.path.join(os.path.dirname(__file__), "rss_subscriptions.json")
        self._subscriptions: Dict[str, List[Dict]] = {}  # {room_id: [{feed_url: str, template_name: str, last_updated: str}]}
        self._templates: Dict[str, Dict] = {}  # {room_id: {"default_template": str, "templates": {name: str, content: str}}}
        self._global_template = self._config.get("module.global_template")
        self._rss_update_interval = self._config.get("module.update_interval")

        self.load_subscriptions()
        asyncio.create_task(self.rss_updater())

    def on_start(self):
        logger.info("RSS Module started")

    def on_stop(self):
        logger.info("RSS Module stopped")

    def load_subscriptions(self):
        try:
            with open(self._subscriptions_file, "r") as f:
                data = json.load(f)
                self._subscriptions = data.get("subscriptions", {})
                self._templates = data.get("templates", {})
        except FileNotFoundError:
            self._subscriptions = {}
            self._templates = {}
        except json.JSONDecodeError:
            logger.warning("Error decoding RSS subscriptions file. Starting with empty data.")
            self._subscriptions = {}
            self._templates = {}

    def save_subscriptions(self):
        data = {"subscriptions": self._subscriptions, "templates": self._templates}
        with open(self._subscriptions_file, "w") as f:
            json.dump(data, f, indent=4)

    async def rss_updater(self):
        logger.debug("RSS polling started")
        while True:
            try:
                await self._poll_once()
            except Exception as e:
                logger.exception(f"Error while polling feeds: {e}")
            await asyncio.sleep(self._rss_update_interval * 60)

    async def _poll_once(self):
        if not self._subscriptions:
            return

        async with aiohttp.ClientSession() as session:
            tasks = []
            for room_id, feeds in self._subscriptions.items():
                for feed_data in feeds:
                    tasks.append(self._process_feed(session, room_id, feed_data))

            # Process feeds concurrently
            await asyncio.gather(*tasks, return_exceptions=True)

    async def _process_feed(self, session: aiohttp.ClientSession, room_id: str, feed_data: dict):
        try:
            feed_info, new_entries = await self._fetch_and_parse_feed(session, feed_data["feed_url"])
            if not new_entries:
                return

            last_updated = datetime.fromisoformat(feed_data.get("last_updated", "1970-01-01T00:00:00"))
            new_entries = [
                entry for entry in new_entries
                if entry["date"] > last_updated
            ]

            if new_entries:
                template = self.get_template(room_id, feed_data.get("template_name"))
                for entry in sorted(new_entries, key=lambda x: x["date"]):
                    message = template.format(
                        feed_url=feed_data["feed_url"],
                        feed_title=feed_info["title"] or feed_data["feed_url"],
                        feed_description=feed_info["description"],
                        title=entry["title"],
                        link=entry["link"],
                        summary=entry["summary"],
                        author=entry.get("author", ""),
                        updated=entry["date"].isoformat()
                    )
                    await self.send_message(room_id, message)

                # Update last_updated timestamp
                feed_data["last_updated"] = max(entry["date"] for entry in new_entries).isoformat()
                self.save_subscriptions()

        except Exception as e:
            logger.warning(f"Failed to process feed {feed_data['feed_url']}: {e}")

    async def _fetch_and_parse_feed(self, session: aiohttp.ClientSession, url: str) -> tuple[
        dict, List[Dict[str, Any]]]:
        headers = {
            "User-Agent": "MatrixBot/1.0 RSS Reader"
        }

        async with session.get(url, headers=headers) as resp:
            content = await resp.read()
            ct = resp.headers["Content-Type"].split(";")[0].strip()

            if ct in ("application/json", "application/feed+json"):
                return await self._parse_json_feed(content)
            else:
                return self._parse_rss_feed(content, url)

    @staticmethod
    def _parse_rss_feed(content: bytes, url: str) -> tuple[dict, List[Dict[str, Any]]]:
        parsed = feedparser.parse(io.BytesIO(content), response_headers={"Content-Location": url})

        feed_info = {
            "title": getattr(parsed.feed, "title", ""),
            "description": getattr(parsed.feed, "description", ""),
            "link": getattr(parsed.feed, "link", "")
        }

        entries = []
        for entry in parsed.entries:
            try:
                date = datetime.fromtimestamp(mktime(entry.get("published_parsed") or entry.get("date_parsed")))
            except (TypeError, ValueError):
                date = datetime.now()

            entries.append({
                "title": getattr(entry, "title", ""),
                "link": getattr(entry, "link", ""),
                "summary": getattr(entry, "description", "").strip(),
                "author": getattr(entry, "author", ""),
                "date": date
            })

        return feed_info, entries

    @staticmethod
    async def _parse_json_feed(content: bytes) -> tuple[dict, List[Dict[str, Any]]]:
        data = json.loads(content)

        feed_info = {
            "title": data.get("title", ""),
            "description": data.get("description", ""),
            "link": data.get("home_page_url", "")
        }

        entries = []
        for item in data.get("items", []):
            try:
                date = datetime.fromisoformat(item["date_published"])
            except (ValueError, KeyError):
                date = datetime.now()

            entries.append({
                "title": item.get("title", ""),
                "link": item.get("url", ""),
                "summary": item.get("summary", item.get("content_html", item.get("content_text", ""))).strip(),
                "author": item.get("author", {}).get("name", ""),
                "date": date
            })

        return feed_info, entries

    def get_template(self, room_id: str, template_name: Optional[str] = None) -> Optional[str]:
        if room_id in self._templates:
            room_templates = self._templates[room_id].get("templates", {})

            if template_name and template_name in room_templates:
                return room_templates[template_name]

            default_template_name = self._templates[room_id].get("default_template")
            if default_template_name and default_template_name in room_templates:
                return room_templates[default_template_name]

        return self._global_template

    @staticmethod
    def _parse_entry_date(entry: Any) -> datetime:
        try:
            if hasattr(entry, "published_parsed") and entry.published_parsed:
                return datetime.fromtimestamp(mktime(entry.published_parsed))
        except (AttributeError, TypeError, ValueError):
            pass

        try:
            if hasattr(entry, "updated_parsed") and entry.updated_parsed:
                return datetime.fromtimestamp(mktime(entry.updated_parsed))
        except (AttributeError, TypeError, ValueError):
            pass

        return datetime.now()

    @BaseModule.command_handler(name="rss", description="RSS feed management.")
    async def rss_command(self, room: MatrixRoom, args: List[str], event: Event):
        await self.send_message(room.room_id, "Use `rss add`, `rss remove`, `rss list`, `rss set` subcommands.")

    @rss_command.subcommand("add", "Add a new RSS feed subscription.", required_level=50)
    async def rss_add(self, room: MatrixRoom, args: List[str], event: Event):
        if len(args) != 1:
            await self.send_message(room.room_id, "Usage: `rss add <feed_url>`")
            return

        feed_url = args[0]

        if room.room_id not in self._subscriptions:
            self._subscriptions[room.room_id] = []

        for sub in self._subscriptions[room.room_id]:
            if sub["feed_url"] == feed_url:
                await self.send_message(room.room_id, f"Feed {feed_url} is already subscribed in this room.")
                return

        self._subscriptions[room.room_id].append({
            "feed_url": feed_url,
            "last_updated": datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        })
        self.save_subscriptions()
        index = len(self._subscriptions[room.room_id]) - 1
        await self.send_message(room.room_id, f"Subscribed to {feed_url} with index {index} using default template.")

    @rss_command.subcommand("remove", "Remove an RSS feed subscription.", required_level=50)
    async def rss_remove(self, room: MatrixRoom, args: List[str], event: Event):
        if len(args) != 1:
            await self.send_message(room.room_id, "Usage: `rss remove <index>`")
            return

        try:
            index = int(args[0])
        except ValueError:
            await self.send_message(room.room_id, "Invalid index. Please provide a valid number.")
            return

        if room.room_id not in self._subscriptions or index >= len(self._subscriptions[room.room_id]):
            await self.send_message(room.room_id, f"No subscription found with index {index} in this room.")
            return

        removed_sub = self._subscriptions[room.room_id].pop(index)
        self.save_subscriptions()
        await self.send_message(room.room_id, f"Unsubscribed from feed {removed_sub['feed_url']} with index {index}.")

    @rss_command.subcommand("list", "List RSS feed subscriptions.")
    async def rss_list(self, room: MatrixRoom, args: List[str], event: Event):
        if room.room_id not in self._subscriptions or not self._subscriptions[room.room_id]:
            await self.send_message(room.room_id, "No RSS subscriptions in this room.")
            return

        message = "<p>RSS subscriptions in this room:</p>\n"
        for index, sub in enumerate(self._subscriptions[room.room_id]):
            template_name = sub.get("template_name", "default")
            message += f"- {index}: {sub['feed_url']} (Template: {template_name})\n"
        await self.send_message(room.room_id, message)

    @BaseModule.command_handler(name="template", description="Template management.")
    async def template_command(self, room: MatrixRoom, args: List[str], event: Event):
        await self.send_message(room.room_id, "Use `template add`, `template set`, `template del`, `template set_default`, `template list` subcommands.")

    @template_command.subcommand("add", "Add a new template.", required_level=50)
    async def template_add(self, room: MatrixRoom, args: List[str], event: RoomMessageText):
        if len(args) < 2:
            await self.send_message(room.room_id, "Usage: `template add <name> <content>`")
            return

        name = args[0]
        content = " ".join(args[1:])

        if room.room_id not in self._templates:
            self._templates[room.room_id] = {"default_template": None, "templates": {}}

        self._templates[room.room_id]["templates"][name] = content
        self.save_subscriptions()
        await self.send_message(room.room_id, f"Template {name} added.")

    @template_command.subcommand("del", "Delete a template.", required_level=50)
    async def template_del(self, room: MatrixRoom, args: List[str], event: Event):
        if len(args) != 1:
            await self.send_message(room.room_id, "Usage: `template del <name>`")
            return

        name = args[0]

        if room.room_id not in self._templates or name not in self._templates[room.room_id]["templates"]:
            await self.send_message(room.room_id, f"No template found with name {name} in this room.")
            return

        if self._templates[room.room_id].get("default_template") == name:
            self._templates[room.room_id]["default_template"] = None

        del self._templates[room.room_id]["templates"][name]
        self.save_subscriptions()
        await self.send_message(room.room_id, f"Template {name} deleted.")

    @template_command.subcommand("set", "Set a template for an RSS feed subscription.", required_level=50)
    async def template_set(self, room: MatrixRoom, args: List[str], event: Event):
        if len(args) < 2:
            await self.send_message(room.room_id, "Usage: `rss set <index> <template_name>`")
            return

        try:
            index = int(args[0])
        except ValueError:
            await self.send_message(room.room_id, "Invalid index. Please provide a valid number.")
            return

        template_name = args[1]

        if room.room_id not in self._subscriptions or index >= len(self._subscriptions[room.room_id]):
            await self.send_message(room.room_id, f"No subscription found with index {index} in this room.")
            return

        if room.room_id not in self._templates or template_name not in self._templates[room.room_id]["templates"]:
            await self.send_message(room.room_id, f"No template found with name {template_name} in this room.")
            return

        self._subscriptions[room.room_id][index]["template_name"] = template_name
        self.save_subscriptions()
        await self.send_message(room.room_id,
                                f"Template updated for subscription with index {index} to {template_name}.")

    @template_command.subcommand("set_default", "Set the default template.", required_level=50)
    async def template_set_default(self, room: MatrixRoom, args: List[str], event: Event):
        if len(args) != 1:
            await self.send_message(room.room_id, "Usage: `template set_default <name>`")
            return

        name = args[0]

        if room.room_id not in self._templates or name not in self._templates[room.room_id]["templates"]:
            await self.send_message(room.room_id, f"No template found with name {name} in this room.")
            return

        self._templates[room.room_id]["default_template"] = name
        self.save_subscriptions()
        await self.send_message(room.room_id, f"Default template set to {name}.")

    @template_command.subcommand("reset_default", "Reset the default template to global default.", required_level=50)
    async def template_reset_default(self, room: MatrixRoom, args: List[str], event: Event):
        if room.room_id not in self._templates:
            await self.send_message(room.room_id, "No templates in this room.")
            return

        if not self._templates[room.room_id].get("default_template"):
            await self.send_message(room.room_id, "No default template is set for this room.")
            return

        self._templates[room.room_id]["default_template"] = None
        self.save_subscriptions()
        await self.send_message(room.room_id, "Default template reset to global default.")

    @template_command.subcommand("list", "List all templates.")
    async def template_list(self, room: MatrixRoom, args: List[str], event: Event):
        if room.room_id not in self._templates or not self._templates[room.room_id]["templates"]:
            await self.send_message(room.room_id, "No templates in this room.")
            return

        message = "<p>Templates in this room:</p>\n"
        for name, content in self._templates[room.room_id]["templates"].items():
            is_default = self._templates[room.room_id].get("default_template") == name
            message += f"- {name}: {content} {'(default)' if is_default else ''}\n"
        await self.send_message(room.room_id, message)
